bl_info = {
    "name": "PivotPro",
    "author": "Jose Conseco",
    "version": (2, 2),
    "blender": (3, 6, 0),
    "location": "3D view",
    "description": "Gives ability to snap pivot",
    "warning": "",
    "wiki_url": "",
    "category": "3D View",
    }


import bpy
from bpy_extras.view3d_utils import region_2d_to_vector_3d, region_2d_to_location_3d
from mathutils import Vector, Matrix, Quaternion


def my_view_orientation(context):
    rv3d = context.space_data.region_3d
    view_inv = rv3d.view_matrix.to_3x3()
    return view_inv.normalized()

def get_orientation_matrix(context):
    orient_type = context.window.scene.transform_orientation_slots[0].type
    mat_x = Vector((1, 0, 0))
    mat_y = Vector((0, 1, 0))
    mat_z = Vector((0, 0, 1))
    custom = context.scene.transform_orientation_slots[0].custom_orientation
    if orient_type == 'CURSOR':
        mat_x = context.scene.cursor.matrix.col[0].xyz
        mat_y = context.scene.cursor.matrix.col[1].xyz
        mat_z = context.scene.cursor.matrix.col[2].xyz
    elif orient_type in {'LOCAL','NORMAL'}:
        obj_mat_w = context.active_object.matrix_world.normalized()
        mat_x = obj_mat_w.col[0].xyz
        mat_y = obj_mat_w.col[1].xyz
        mat_z = obj_mat_w.col[2].xyz
    elif orient_type == 'VIEW':
        view_ma = my_view_orientation(context)
        mat_x = view_ma[0].xyz
        mat_y = view_ma[1].xyz
        mat_z = view_ma[2].xyz
    elif custom:
        mat_x = custom.matrix[0].xyz
        mat_y = custom.matrix[1].xyz
        mat_z = custom.matrix[2].xyz

    return mat_x, mat_y, mat_z



class MoveCursorGizmoGroup(bpy.types.GizmoGroup):
    bl_idname = "CURSOR_GT_cursor_move"
    bl_label = "Gizmo for 3d Cursor"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'WINDOW'
    bl_options = {'3D', 'SHOW_MODAL_ALL', 'PERSISTENT'}

    # Helper functions
    @staticmethod
    def my_target_operator(context):
        wm = context.window_manager
        op = wm.operators[-1] if wm.operators else None
        if isinstance(op, PP_OT_PivotTransform):
            return op
        return None

    @classmethod
    def poll(cls, context):
        return context.scene.pivot_pro_enabled and (context.scene.tool_settings.transform_pivot_point == 'CURSOR' or not context.selected_objects)

        # op = cls.my_target_operator(context)
        # if op is None:
        #     wm = context.window_manager
        #     wm.gizmo_group_type_unlink_delayed(MoveCursorGizmoGroup.bl_idname)
        #     return False
        # return True


    def setup(self, context):
        ui = bpy.context.preferences.themes[0].user_interface
        colorXYZ = [ui.axis_x, ui.axis_y, ui.axis_z]

        #move x
        arrow_x = self.gizmos.new("GIZMO_GT_arrow_3d")
        op = arrow_x.target_set_operator(PP_OT_PivotTransform.bl_idname)
        op.axis = 'X'
        self.move_x = arrow_x

        #move y
        arrow_y = self.gizmos.new("GIZMO_GT_arrow_3d")
        op = arrow_y.target_set_operator(PP_OT_PivotTransform.bl_idname)
        op.axis = 'Y'
        self.move_y = arrow_y

        #move z
        arrow_z = self.gizmos.new("GIZMO_GT_arrow_3d")
        op = arrow_z.target_set_operator(PP_OT_PivotTransform.bl_idname)
        op.axis = 'Z'
        self.move_z = arrow_z

        move_3d = self.gizmos.new("GIZMO_GT_move_3d")
        move_3d.draw_options = {"ALIGN_VIEW"}
        move_3d.line_width = 4
        op = move_3d.target_set_operator(PP_OT_PivotTransform.bl_idname)
        op.axis = 'NONE'
        self.move_3d = move_3d

        for i, g in enumerate(self.gizmos):
            g.color = colorXYZ[i % 3]
            g.alpha_highlight = 1.0
            g.color_highlight = 1.0, 1.0, 1.0
            g.alpha = 0.8
            g.use_draw_modal = True
            g.scale_basis = 1.0

        move_3d.scale_basis = 0.2


    def draw_prepare(self, context):
        gizmo_loc = context.scene.cursor.location
        mat_x, mat_y, mat_z = get_orientation_matrix(context)

        m_x = Matrix.Identity(4)
        m_x.col[0].xyz = mat_z
        m_x.col[1].xyz = mat_y
        m_x.col[2].xyz = mat_x
        m_x.col[3].xyz = gizmo_loc
        self.move_x.matrix_basis = m_x
        self.move_3d.matrix_basis = m_x

        m_y = Matrix.Identity(4)
        m_y.col[0].xyz = mat_x
        m_y.col[1].xyz = mat_z
        m_y.col[2].xyz = mat_y
        m_y.col[3].xyz = gizmo_loc
        self.move_y.matrix_basis = m_y

        m_z = Matrix.Identity(4)
        m_z.col[0].xyz = mat_x
        m_z.col[1].xyz = mat_y
        m_z.col[2].xyz = mat_z
        m_z.col[3].xyz = gizmo_loc
        self.move_z.matrix_basis = m_z


class PP_OT_PivotTransform(bpy.types.Operator):
    bl_idname = "object.cursor_pivot_transform"
    bl_label = "Transform objects with 3Dcursor pivot"
    bl_options = {"REGISTER"}

    operator: bpy.props.StringProperty("")
    __slots__ = ['constraints', 'pivot_point', 'orientation_matrix', 'finished', 'init_co', '_timer']

    axis: bpy.props.EnumProperty(name='Axis', description='Axis', items=[
        ('X', 'x', '', '', 0),
        ('Y', 'y', '', '', 1),
        ('Z', 'z', '', '', 2),
        ('NONE', 'None', '', '', 3)],
        default='NONE')


    def modal(self, context, event):
        wm = context.window_manager
        if event.type == 'LEFTMOUSE':  # ? never used
            # if context.space_data.type == 'VIEW_3D':
            #     context.window_manager.gizmo_group_type_ensure(MoveCursorGizmoGroup.bl_idname)
            wm.event_timer_remove(self._timer)
            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}: #? never used
            # print('esc')
            self.finished = True
            wm.event_timer_remove(self._timer)
            return {'CANCELLED'}

        elif event.type == 'TIMER' and context.selected_objects: #update selected objs on timer if any
            # print('timers')
            context.view_layer.update()  # TO GET UPDATE CURSORLOC
            dif = context.scene.cursor.location - self.init_co
            for obj in context.selected_objects:
                obj.matrix_world.translation += dif
            self.init_co = context.scene.cursor.location.copy()

        elif event.type == 'MOUSEMOVE':
            # print('move')
            if self.finished:
                #* old way - by gettin last operator (bpy.transform()) props
                # orient_type = context.scene.transform_orientation_slots[0].type
                #cos this operator pust -2 ops on history stack:   [..., transfrom(only_curosr), self] on stack
                # last_translate_props = wm.operator_properties_last(wm.operators[-1].bl_idname)
                # last_translate_props['cursor_transform'] = False
                # kwargs = {key: getattr(last_translate_props, key) for key, val in last_translate_props.items()}
                # kwargs['constraint_axis'] = self.constraints
                # const_mat = Matrix.Identity(3)
                # return bpy.ops.transform.translate('EXEC_DEFAULT', value=kwargs['value'], constraint_axis=self.constraints,  cursor_transform=False,  # release_confirm=True,
                #                                    orient_type=orient_type, orient_matrix_type='GLOBAL', orient_matrix=const_mat, snap=False)
                # bpy.ops.transform.translate(**kwargs)

                context.view_layer.update()  # TO GET UPDATE CURSORLOC
                dif = context.scene.cursor.location - self.init_co
                for obj in context.selected_objects:
                    obj.matrix_world.translation += dif
                wm.event_timer_remove(self._timer)
                # print('fins')
                return {'FINISHED'}
            self.finished = True
            return bpy.ops.transform.translate('INVOKE_DEFAULT', constraint_axis=self.constraints,  cursor_transform=True,  # release_confirm=True,
                                               orient_type='GLOBAL', orient_matrix_type='GLOBAL', orient_matrix=self.orientation_matrix)  # ,snap=event.ctrl
            # context.area.tag_redraw()

        return {'PASS_THROUGH'}

    def invoke(self, context, event):
        mat_x, mat_y, mat_z = get_orientation_matrix(context)
        orient_mat = Matrix.Identity(3)
        orient_mat.col[0].xyz = mat_x
        orient_mat.col[1].xyz = mat_y
        orient_mat.col[2].xyz = mat_z
        self.orientation_matrix = orient_mat

        self.init_co = context.scene.cursor.location.copy()

        wm = context.window_manager
        self._timer = wm.event_timer_add(0.1, window=context.window)

        #? not Needed
        # if context.space_data.type == 'VIEW_3D':
        #     context.window_manager.gizmo_group_type_ensure(MoveCursorGizmoGroup.bl_idname)
        self.constraints = (self.axis == 'X', self.axis == 'Y', self.axis == 'Z')
        self.finished = False
        if context.scene.pivot_pro_enabled:
            if context.scene.tool_settings.transform_pivot_point == 'CURSOR' or not context.selected_objects: # if no sel objs - then move pivot/cursor
                context.window_manager.modal_handler_add(self)
                return {'RUNNING_MODAL'}
            else:
                return {'PASS_THROUGH'}  # this way proper undo stack with translate
        else:
            return {'PASS_THROUGH'}  # this way proper undo stack with translate


def matrix_decompose(mat):
    ''' returns active_obj_mat_loc, active_obj_mat_rot, active_obj_mat_sca
        reconstruct by loc @ rotQuat @ scale '''

    loc, rotQuat, scale = mat.decompose()
    mat_loc = Matrix.Translation(loc)
    mat_rot = rotQuat.to_matrix().to_4x4()
    mat_sca = Matrix()
    for i in range(3):
        mat_sca[i][i] = scale[i]

    return mat_loc, mat_rot, mat_sca

class PP_OT_PivotRestRot(bpy.types.Operator):
    bl_idname = "pivot.reset_rotation"
    bl_label = "Rest Pivot Rotation"
    bl_description = "Overrides build-in rotation clear"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def invoke(self, context, event):
        if context.scene.pivot_pro_enabled:
            if len(context.selected_objects) == 0 or (len(context.selected_objects) > 0 and context.scene.tool_settings.transform_pivot_point == 'CURSOR'):
                return self.execute(context)
            else:  # this way proper undo stack with rotation_clear
                # return bpy.ops.object.rotation_clear('INVOKE_DEFAULT')
                return {'PASS_THROUGH'}  #run default clear rot
        else:
            return {'PASS_THROUGH'}  # run default clear rto

    def execute(self, context):
        if len(context.selected_objects) > 0 and context.scene.tool_settings.transform_pivot_point == 'CURSOR':
            mat_loc, mat_rot, mat_sca = matrix_decompose(context.scene.cursor.matrix)
            for obj in context.selected_objects:
                obj.matrix_world = mat_loc @ mat_rot.inverted() @ mat_loc.inverted() @ obj.matrix_world
        back = context.scene.cursor.rotation_mode
        context.scene.cursor.rotation_mode = 'QUATERNION'
        context.scene.cursor.rotation_quaternion = Quaternion((1.0, 0.0, 0.0, 0.0))
        context.scene.cursor.rotation_mode = back
        return {"FINISHED"}



class PP_OT_PivotRestLoc(bpy.types.Operator):
    bl_idname = "pivot.reset_location"
    bl_label = "Rest Pivot Location"
    bl_description = "Overrides build-in translation clear"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return context.active_object

    def invoke(self, context, event):
        if context.scene.pivot_pro_enabled:
            if len(context.selected_objects) == 0 or (len(context.selected_objects) > 0 and context.scene.tool_settings.transform_pivot_point == 'CURSOR'):
                return self.execute(context)
            else:  # this way proper undo stack with rotation_clear
                # return bpy.ops.object.location_clear('INVOKE_DEFAULT')
                return {'PASS_THROUGH'}  # run default clear rot
        else:
            return {'PASS_THROUGH'}  # run default clear rto

    def execute(self, context):
        if len(context.selected_objects) > 0 and context.scene.tool_settings.transform_pivot_point == 'CURSOR':
            cur_loc = context.scene.cursor.location
            for obj in context.selected_objects:
                obj.matrix_world.translation -= cur_loc
        context.scene.cursor.location = Vector((0,0,0))
        return {"FINISHED"}


addon_keymaps = [] #put on out of register()
def RegisterHotkeys(enable):
    if enable:
        wm = bpy.context.window_manager
        km = wm.keyconfigs.addon.keymaps.new(name = 'Object Mode', space_type = 'EMPTY')
        kmi = km.keymap_items.new(PP_OT_PivotTransform.bl_idname, 'LEFTMOUSE', 'DOUBLE_CLICK')
        kmi.properties.axis = 'NONE'
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new(PP_OT_PivotTransform.bl_idname, 'G', 'PRESS')  # , shift=True
        kmi.properties.operator = "Translate"
        kmi.properties.axis = "NONE"
        addon_keymaps.append((km, kmi))

        # kmi = km.keymap_items.new(PP_OT_PivotTransform.bl_idname, 'R', 'PRESS')
        # kmi.properties.operator = "Rotate"
        # kmi.properties.axis = "NONE"
        # addon_keymaps.append((km, kmi))

        # kmi = km.keymap_items.new(PP_OT_PivotTransform.bl_idname, 'S', 'PRESS')
        # kmi.properties.operator = "Scale"
        # kmi.properties.axis = "NONE"
        # addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new(PP_OT_PivotRestRot.bl_idname, 'R', 'PRESS', alt=True)  # , shift=True
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new(PP_OT_PivotRestLoc.bl_idname, 'G', 'PRESS', alt=True)  # , shift=True
        addon_keymaps.append((km, kmi))
    else:
        for km, kmi in addon_keymaps:
            km.keymap_items.remove(kmi)
        addon_keymaps.clear()


# Object that will store the handle to the msgbus subscription
subscription_owner = object()
# def msgbus_callback(*args):  # bpy.types.ToolSettings
#         print("msgbus_callback", args)
#         if args[0] == 'CURSOR': #pivot is set to cursor - then use hotkeys
#         RegisterHotkeys(bpy.context.scene.pivot_pro_enabled)

def ppro_prop_change(self, context):  # just when enabling disablig button 'pivotPro'
    RegisterHotkeys(self.pivot_pro_enabled)

    #enable ol disable
    # if self.pivot_pro_enabled:
    #     subscribe_to = context.scene.tool_settings.path_resolve("transform_pivot_point", False)
    #     args = (context.scene.tool_settings.transform_pivot_point)
    #     bpy.msgbus.subscribe_rna(key=subscribe_to, owner=subscription_owner, args=(1, 2, 3), notify=msgbus_callback)
    # else:
    #     bpy.msgbus.clear_by_owner(subscription_owner)


classes = (
    MoveCursorGizmoGroup,
    PP_OT_PivotTransform,
    PP_OT_PivotRestRot,
    PP_OT_PivotRestLoc,
)


def gizmo_button(self, context):
    layout = self.layout
    layout.prop(context.scene, "pivot_pro_enabled", text='PivotPro', icon='OUTLINER_OB_EMPTY')

def register():
    print('Pivot Pro Register')
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)

    bpy.types.Scene.pivot_pro_enabled = bpy.props.BoolProperty(name="Enable PivotPro", description="Turns on/off pivot", default=False, update=ppro_prop_change)
    bpy.types.VIEW3D_PT_gizmo_display.append(gizmo_button)

    try:
        RegisterHotkeys(bpy.context.scene.pivot_pro_enabled)
    except:
        pass



def unregister():
    print('Pivot Pro unregister')

    bpy.types.VIEW3D_PT_gizmo_display.remove(gizmo_button)
    RegisterHotkeys(bpy.context.scene.pivot_pro_enabled)

    bpy.context.scene.pivot_pro_enabled = False
    del bpy.types.Scene.pivot_pro_enabled

    from bpy.utils import unregister_class
    for cls in reversed(classes):
        unregister_class(cls)



